@@ -1,4 +1,5 @@
 #include <stdio.h>
+#include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -16,32 +17,54 @@
 #define HEAVY_CHECK_COUNT	10
 
 struct emc_clariion_checker_context {
+	int fd;
 	int run_count;
 	char wwn[16];
 	unsigned wwn_set;
 };
 
-int emc_clariion(char *devnode, char *msg, void *context)
+int emc_clariion(char *devt, char *msg, void **context)
 {
 	unsigned char sense_buffer[128];
 	unsigned char inqCmdBlk[INQUIRY_CMDLEN] = {INQUIRY_CMD, 0, 0xC0, 0,
 						sizeof(sense_buffer), 0};
 	struct sg_io_hdr io_hdr;
-	int fd;
-	struct emc_clariion_checker_context * ctxt;
+	struct emc_clariion_checker_context * ctxt = NULL;
+	int ret;
 
-	if (context != NULL) {
-		ctxt = (struct emc_clariion_checker_context *)context;
-		ctxt->run_count++;
+	/*
+	 * caller passed in a context : use its address
+	 */
+	if (context)
+		ctxt = (struct emc_clariion_checker_context *) (*context);
+
+	/*
+	 * passed in context is uninitialized or volatile context :
+	 * initialize it
+	 */
+	if (!ctxt) {
+		ctxt = malloc(sizeof(struct emc_clariion_checker_context *));
+		memset(ctxt, 0, sizeof(struct emc_clariion_checker_context *));
 	}
+	if (!ctxt) {
+		MSG("cannot allocate context");
+		return -1;
+	}
+	ctxt->run_count++;
 
 	if (ctxt->run_count % HEAVY_CHECK_COUNT) {
 		ctxt->run_count = 0;
 		/* do stuff */
 	}
 
-	fd = open (devnode, O_RDONLY);
-
+	if (!ctxt->fd) {
+		if (devnode(CREATE_NODE, devt)) {
+			ret = -1;
+			goto out;
+		}
+		ctxt->fd = devnode(OPEN_NODE, devt);
+		devnode(UNLINK_NODE, devt);
+	}
 	memset(&io_hdr, 0, sizeof (struct sg_io_hdr));
 	io_hdr.interface_id = 'S';
 	io_hdr.cmd_len = sizeof (inqCmdBlk);
@@ -51,22 +74,21 @@ int emc_clariion(char *devnode, char *msg, void *context)
 	io_hdr.sbp = sense_buffer;
 	io_hdr.timeout = 20000;
 	io_hdr.pack_id = 0;
-	if (ioctl(fd, SG_IO, &io_hdr) < 0) {
-		close (fd);
+	if (ioctl(ctxt->fd, SG_IO, &io_hdr) < 0) {
 		MSG("emc_clariion_checker: sending query command failed");
-		return PATH_DOWN;
+		ret = PATH_DOWN;
+		goto out;
 	}
 	if (io_hdr.info & SG_INFO_OK_MASK) {
-		close (fd);
 		MSG("emc_clariion_checker: query command indicates error");
-		return PATH_DOWN;
+		ret = PATH_DOWN;
+		goto out;
 	}
-	close (fd);
-
 	if (/* Verify the code page - right page & revision */
 	    sense_buffer[1] != 0xc0 || sense_buffer[9] != 0x00) {
 		MSG("emc_clariion_checker: Path unit report page in unknown format");
-		return PATH_DOWN;
+		ret = PATH_DOWN;
+		goto out;
 	}
 
 	if ( /* Effective initiator type */
@@ -76,13 +98,15 @@ int emc_clariion(char *devnode, char *msg, void *context)
 		/* Arraycommpath should be set to 1 */
 		|| (sense_buffer[30] & 0x04) != 0x04) {
 		MSG("emc_clariion_checker: Path not correctly configured for failover");
-		return PATH_DOWN;
+		ret = PATH_DOWN;
+		goto out;
 	}
 
 	if ( /* LUN operations should indicate normal operations */
 		sense_buffer[48] != 0x00) {
 		MSG("emc_clariion_checker: Path not available for normal operations");
-		return PATH_SHAKY;
+		ret = PATH_SHAKY;
+		goto out;
 	}
 
 #if 0
@@ -90,7 +114,8 @@ int emc_clariion(char *devnode, char *msg, void *context)
 	 * _would_ bind the path */
 	if ( /* LUN should at least be bound somewhere */
 		sense_buffer[4] != 0x00) {
-		return PATH_UP;
+		ret = PATH_UP;
+		goto out;
 	}
 #endif	
 	
@@ -103,7 +128,8 @@ int emc_clariion(char *devnode, char *msg, void *context)
 	if (ctxt->wwn_set) {
 		if (!memcmp(ctxt->wwn, &sense_buffer[10], 16)) {
 			MSG("emc_clariion_checker: Logical Unit WWN has changed!");
-			return PATH_DOWN;
+			ret = PATH_DOWN;
+			goto out;
 		}
 	} else {
 		memcpy(ctxt->wwn, &sense_buffer[10], 16);
@@ -112,6 +138,15 @@ int emc_clariion(char *devnode, char *msg, void *context)
 	
 	
 	MSG("emc_clariion_checker: Path healthy");
-        return PATH_UP;
-
+        ret = PATH_UP;
+out:
+	/*
+	 * caller told us he doesn't want to keep the context :
+	 * free it
+	 */
+	if (!context) {
+		close(ctxt->fd);
+		free(ctxt);
+	}
+	return(ret);
 }
