@@ -36,7 +36,6 @@
 #include <checkers.h>
 #include <path_state.h>
 #include <safe_printf.h>
-#include <callout.h>
 
 #include "main.h"
 #include "devinfo.h"
@@ -44,269 +43,11 @@
 #include "pgpolicies.h"
 #include "dict.h"
 #include "debug.h"
+#include "dmparser.h"
+#include "propsel.h"
 
 /* helpers */
-#define argis(x) if (0 == strcmp (x, argv[i]))
-#define MATCH(x,y) 0 == strncmp (x, y, strlen(y))
-
-static struct hwentry *
-find_hw (char * vendor, char * product)
-{
-	int i;
-	struct hwentry * hwe;
-
-	vector_foreach_slot (conf->hwtable, hwe, i)
-		if (hwe->vendor && hwe->product &&
-		    strcmp(hwe->vendor, vendor) == 0 &&
-		    strcmp(hwe->product, product) == 0)
-			return hwe;
-	return NULL;
-}
-
-static struct mpentry *
-find_mp (char * wwid)
-{
-	int i;
-	struct mpentry * mpe;
-
-	vector_foreach_slot (conf->mptable, mpe, i)
-                if (mpe->wwid && strcmp(mpe->wwid, wwid) == 0)
-			return mpe;
-
-	return NULL;
-}
-
-/*
- * selectors :
- * traverse the configuration layers from most specific to most generic
- * stop at first explicit setting found
- */
-static int
-select_pgpolicy (struct multipath * mp)
-{
-	int i;
-	struct mpentry * mpe = NULL;
-	struct hwentry * hwe = NULL;
-	struct path * pp;
-	char pgpolicy_name[POLICY_NAME_SIZE];
-
-	pp = VECTOR_SLOT(mp->paths, 0);
-
-	if (conf->pgpolicy_flag > 0) {
-		mp->pgpolicy = conf->pgpolicy_flag;
-		get_pgpolicy_name(pgpolicy_name, mp->pgpolicy);
-		dbg("pgpolicy = %s (cmd line flag)", pgpolicy_name);
-		return 0;
-	}
-	vector_foreach_slot (conf->mptable, mpe, i) {
-		if (mpe->wwid && strcmp(mpe->wwid, mp->wwid) == 0 &&
-		    mpe->pgpolicy > 0) {
-			mp->pgpolicy = mpe->pgpolicy;
-			get_pgpolicy_name(pgpolicy_name, mp->pgpolicy);
-			dbg("pgpolicy = %s (LUN setting)", pgpolicy_name);
-			return 0;
-		}
-	}
-	vector_foreach_slot (conf->hwtable, hwe, i) {
-		if (hwe->vendor && hwe->product &&
-		    MATCH (pp->vendor_id, hwe->vendor) &&
-		    MATCH (pp->product_id, hwe->product) &&
-		    hwe->pgpolicy > 0) {
-			mp->pgpolicy = hwe->pgpolicy;
-			get_pgpolicy_name(pgpolicy_name, mp->pgpolicy);
-			dbg("pgpolicy = %s (controler setting)", pgpolicy_name);
-			return 0;
-		}
-	}
-	if (conf->default_pgpolicy > 0) {
-		mp->pgpolicy = conf->default_pgpolicy;
-		get_pgpolicy_name(pgpolicy_name, mp->pgpolicy);
-		dbg("pgpolicy = %s (config file default)", pgpolicy_name);
-		return 0;
-	}
-	mp->pgpolicy = FAILOVER;
-	get_pgpolicy_name(pgpolicy_name, FAILOVER);
-	dbg("pgpolicy = %s (internal default)", pgpolicy_name);
-	return 0;
-}
-
-static int
-select_selector (struct multipath * mp)
-{
-	int i;
-	struct mpentry * mpe = NULL;
-	struct hwentry * hwe = NULL;
-	struct path * pp;
-
-	pp = VECTOR_SLOT(mp->paths, 0);
-
-	mpe = find_mp(mp->wwid);
-
-	if (mpe && mpe->selector) {
-		mp->selector = mpe->selector;
-		mp->selector_args = mpe->selector_args;
-		dbg("selector = %s (LUN setting)", mp->selector);
-		dbg("selector_args = %s (LUN setting)", mp->selector_args);
-		return 0;
-	}
-	hwe = find_hw(pp->vendor_id, pp->product_id);
-
-	if (hwe && hwe->selector) {
-		mp->selector = hwe->selector;
-		mp->selector_args = hwe->selector_args;
-		dbg("selector = %s (controler setting)", mp->selector);
-		dbg("selector_args = %s (controler setting)", mp->selector_args);
-		return 0;
-	}
-	mp->selector = conf->default_selector;
-	mp->selector_args = conf->default_selector_args;
-	dbg("selector = %s (internal default)", mp->selector);
-	dbg("selector_args = %s (internal default)", mp->selector_args);
-	return 0;
-}
-
-static int
-select_features (struct multipath * mp)
-{
-	int i;
-	struct hwentry * hwe = NULL;
-	struct path * pp;
-
-	pp = VECTOR_SLOT(mp->paths, 0);
-	hwe = find_hw(pp->vendor_id, pp->product_id);
-
-	if (hwe && hwe->features) {
-		mp->features = hwe->features;
-		dbg("features = %s (controler setting)", mp->features);
-		return 0;
-	}
-	mp->features = conf->default_features;
-	dbg("features = %s (internal default)", mp->features);
-	return 0;
-}
-
-static int
-select_hwhandler (struct multipath * mp)
-{
-	int i;
-	struct hwentry * hwe = NULL;
-	struct path * pp;
-
-        pp = VECTOR_SLOT(mp->paths, 0);
-	hwe = find_hw(pp->vendor_id, pp->product_id);
-
-	if (hwe && hwe->hwhandler) {
-		mp->hwhandler = hwe->hwhandler;
-		dbg("hwhandler = %s (controler setting)", mp->hwhandler);
-		return 0;
-	}
-	mp->hwhandler = conf->default_hwhandler;
-	dbg("hwhandler = %s (internal default)", mp->hwhandler);
-	return 0;
-}
-
-static int
-select_checkfn(struct path *pp)
-{
-	char checker_name[CHECKER_NAME_SIZE];
-	int i;
-	struct hwentry * hwe = NULL;
-
-	hwe = find_hw(pp->vendor_id, pp->product_id);
-
-	if (hwe && hwe->checker_index > 0) {
-		get_checker_name(checker_name, hwe->checker_index);
-		dbg("path checker = %s (controler setting)", checker_name);
-		pp->checkfn = get_checker_addr(hwe->checker_index);
-		return 0;
-	}
-	pp->checkfn = &readsector0;
-	get_checker_name(checker_name, READSECTOR0);
-	dbg("path checker = %s (internal default)", checker_name);
-	return 0;
-}
-
-static int
-select_getuid (struct path * pp)
-{
-	int i;
-	struct hwentry * hwe = NULL;
-
-	hwe = find_hw(pp->vendor_id, pp->product_id);
-
-	if (hwe && hwe->getuid) {
-		pp->getuid = hwe->getuid;
-		dbg("getuid = %s (controler setting)", pp->getuid);
-		return 0;
-	}
-	pp->getuid = conf->default_getuid;
-	dbg("getuid = %s (internal default)", pp->getuid);
-	return 0;
-}
-
-static int
-select_getprio (struct path * pp)
-{
-	int i;
-	struct hwentry * hwe = NULL;
-
-	hwe = find_hw(pp->vendor_id, pp->product_id);
-
-	if (hwe && hwe->getprio) {
-		pp->getprio = hwe->getprio;
-		dbg("getprio = %s (controler setting)", pp->getprio);
-		return 0;
-	}
-	pp->getprio = conf->default_getprio;
-	dbg("getprio = %s (internal default)", pp->getprio);
-	return 0;
-}
-
-static int
-devt2devname (char *devname, char *devt)
-{
-	struct sysfs_directory * sdir;
-	struct sysfs_directory * devp;
-	char sysfs_path[FILE_NAME_SIZE];
-	char block_path[FILE_NAME_SIZE];
-	char attr_path[FILE_NAME_SIZE];
-	char attr_value[16];
-
-	if (sysfs_get_mnt_path(sysfs_path, FILE_NAME_SIZE)) {
-		fprintf(stderr, "-D feature available with sysfs only\n");
-		exit(1);
-	}
-		
-	if(safe_sprintf(block_path, "%s/block", sysfs_path)) {
-		fprintf(stderr, "block_path too small\n");
-		exit(1);
-	}
-	sdir = sysfs_open_directory(block_path);
-	sysfs_read_directory(sdir);
-
-	dlist_for_each_data (sdir->subdirs, devp, struct sysfs_directory) {
-		if(safe_sprintf(attr_path, "%s/%s/dev",
-				block_path, devp->name)) {
-			fprintf(stderr, "attr_path too small\n");
-			exit(1);
-		}
-		sysfs_read_attribute_value(attr_path, attr_value,
-					   sizeof(attr_value));
-
-		if (!strncmp(attr_value, devt, strlen(devt))) {
-			if(safe_sprintf(attr_path, "%s/%s",
-					block_path, devp->name)) {
-				fprintf(stderr, "attr_path too small\n");
-				exit(1);
-			}
-			sysfs_get_name_from_path(attr_path, devname,
-						 FILE_NAME_SIZE);
-			break;
-		}
-	}
-	sysfs_close_directory(sdir);
-	return 0;
-}
+#define argis(x) if (0 == strcmp(x, argv[i]))
 
 static int
 filepresent (char * run) {
@@ -329,153 +70,6 @@ blacklist (char * dev) {
 	return 0;
 }
 
-static char *
-apply_format (char * string, int maxsize, struct path * pp)
-{
-	char * pos;
-	char * dst;
-	char * p;
-	char c;
-	int len;
-	int free;
-
-	dst = zalloc(maxsize);
-
-	if (!dst)
-		return NULL;
-
-	p = dst;
-	pos = strchr(string, '%');
-	free = maxsize;
-
-	if (!pos)
-		return string;
-
-	len = (int) (pos - string) + 1;
-	free -= len;
-
-	if (free < 2)
-		return NULL;
-
-	snprintf(p, len, "%s", string);
-	p += len - 1;
-	pos++;
-
-	switch (*pos) {
-	case 'n':
-		len = strlen(pp->dev) + 1;
-		free -= len;
-
-		if (free < 2)
-			return NULL;
-
-		snprintf(p, len, "%s", pp->dev);
-		p += len - 1;
-		break;
-	case 'd':
-		len = strlen(pp->dev_t) + 1;
-		free -= len;
-
-		if (free < 2)
-			return NULL;
-
-		snprintf(p, len, "%s", pp->dev_t);
-		p += len - 1;
-		break;
-	default:
-		break;
-	}
-	pos++;
-
-	if (!*pos)
-		return dst;
-
-	len = strlen(pos) + 1;
-	free -= len;
-
-	if (free < 2)
-		return NULL;
-
-	snprintf(p, len, "%s", pos);
-	dbg("reformated callout = %s", dst);
-	return dst;
-}
-
-static int
-devinfo (struct path *pp)
-{
-	int i;
-	struct hwentry * hwe;
-	char * buff;
-	char prio[16];
-
-	dbg("===== path %s =====", pp->dev);
-
-	/*
-	 * fetch info available in sysfs
-	 */
-	if (sysfs_devinfo(pp))
-		return 1;
-
-	/*
-	 * then those not available through sysfs
-	 */
-	get_serial(pp->serial, pp->dev_t);
-	dbg("serial = %s", pp->serial);
-	pp->claimed = get_claimed(pp->dev_t);
-	dbg("claimed = %i", pp->claimed);
-
-	/*
-	 * get path state, no message collection, no context
-	 */
-	select_checkfn(pp);
-	pp->state = pp->checkfn(pp->dev_t, NULL, NULL);
-	dbg("state = %i", pp->state);
-	
-	/*
-	 * get path prio
-	 */
-	select_getprio(pp);
-	buff = apply_format(pp->getprio, CALLOUT_MAX_SIZE, pp);
-
-	if (!buff)
-		pp->priority = 1;
-	else if (execute_program(buff, prio, 16)) {
-		dbg("error calling out %s", buff);
-		pp->priority = 1;
-	} else
-		pp->priority = atoi(prio);
-
-	dbg("prio = %u", pp->priority);
-
-	/*
-	 * get path uid
-	 */
-	select_getuid(pp);
-
-	buff = apply_format(pp->getuid, CALLOUT_MAX_SIZE, pp);
-
-	if (!buff)
-		goto fallback;
-
-	if (execute_program(buff, pp->wwid, WWID_SIZE) == 0) {
-		dbg("uid = %s (callout)", pp->wwid);
-		return 0;
-	}
-
-	fallback:
-	if (!get_evpd_wwid(pp->dev_t, pp->wwid)) {
-		dbg("uid = %s (internal getuid)", pp->wwid);
-		return 0;
-	}
-	/*
-	 * no wwid : blank for safety
-	 */
-	dbg("uid = 0x0 (unable to fetch)");
-	memset(pp->wwid, 0, WWID_SIZE);
-	return 1;
-}
-
 static int
 get_pathvec_sysfs (vector pathvec)
 {
@@ -502,6 +96,7 @@ get_pathvec_sysfs (vector pathvec)
 	 * that relate to the device pointed by conf->dev
 	 */
 	if (conf->dev && filepresent(conf->dev)) {
+		dbg("limited scope = %s", conf->dev);
 		curpath = zalloc(sizeof (struct path));
 		basename(conf->dev, curpath->dev);
 
@@ -567,20 +162,11 @@ get_pathvec_sysfs (vector pathvec)
 			free (curpath);
 			continue;
 		}
-
 		if (memcmp(empty_buff, refwwid, WWID_SIZE) != 0 && 
 		    memcmp(curpath->wwid, refwwid, WWID_SIZE) != 0) {
 			free(curpath);
 			continue;
 		}
-
-		basename(linkp->target, buff);
-		sscanf(buff, "%i:%i:%i:%i",
-			&curpath->sg_id.host_no,
-			&curpath->sg_id.channel,
-			&curpath->sg_id.scsi_id,
-			&curpath->sg_id.lun);
-
 		vector_alloc_slot(pathvec);
 		vector_set_slot(pathvec, curpath);
 	}
@@ -589,27 +175,30 @@ get_pathvec_sysfs (vector pathvec)
 }
 
 /*
- * print_path style
+ * print_path styles
  */
-#define ALL	0
-#define SHORT	1
+#define PRINT_PATH_ALL		0
+#define PRINT_PATH_SHORT	1
 
 static void
 print_path (struct path * pp, int style)
 {
-	if (style != SHORT && pp->wwid)
+	if (style != PRINT_PATH_SHORT && pp->wwid)
 		printf ("%s ", pp->wwid);
 	else
-		printf (" \\_");
+		printf ("  \\_ ");
 
-	printf("(%i %i %i %i) ",
+	printf("%i:%i:%i:%i ",
 	       pp->sg_id.host_no,
 	       pp->sg_id.channel,
 	       pp->sg_id.scsi_id,
 	       pp->sg_id.lun);
 
 	if (pp->dev)
-		printf("%s ", pp->dev);
+		printf("%-4s ", pp->dev);
+
+	if (pp->dev_t)
+		printf("%-7s ", pp->dev_t);
 
 	switch (pp->state) {
 	case PATH_UP:
@@ -625,66 +214,135 @@ print_path (struct path * pp, int style)
 		printf("[undef ]");
 		break;
 	}
-	if (pp->dev_t)
-		printf("[%s]", pp->dev_t);
-
+	switch (pp->dmstate) {
+	case PSTATE_ACTIVE:
+		printf("[active]");
+		break;
+	case PSTATE_FAILED:
+		printf("[failed]");
+		break;
+	default:
+		break;
+	}
 	if (pp->claimed)
 		printf("[claimed]");
 
-	if (style != SHORT && pp->product_id)
+	if (style != PRINT_PATH_SHORT && pp->product_id)
 		printf("[%.16s]", pp->product_id);
 
 	fprintf(stdout, "\n");
 }
 
+#if DEBUG
 static void
-print_all_path (vector pathvec)
+print_map (struct multipath * mpp)
 {
-	int k;
+	if (mpp->size && mpp->params)
+		printf("0 %lu %s %s\n",
+			 mpp->size, DM_TARGET, mpp->params);
+	return;
+}
+
+static void
+print_all_paths (vector pathvec)
+{
+	int i;
 	char empty_buff[WWID_SIZE];
 	struct path * pp;
 
 	/* initialize a cmp 0-filled buffer */
-	memset (empty_buff, 0, WWID_SIZE);
+	memset(empty_buff, 0, WWID_SIZE);
 
-	fprintf (stdout, "#\n# all paths :\n#\n");
-
-	vector_foreach_slot (pathvec, pp, k) {
+	vector_foreach_slot (pathvec, pp, i) {
 		/* leave out paths with incomplete devinfo */
-		if (memcmp (empty_buff, pp->wwid, WWID_SIZE) == 0)
+		if (memcmp(empty_buff, pp->wwid, WWID_SIZE) == 0)
 			continue;
 
-		print_path (pp, ALL);
+		print_path(pp, PRINT_PATH_ALL);
 	}
 }
 
+static void
+print_all_maps (vector mp)
+{
+	int i;
+	struct multipath * mpp;
+
+	vector_foreach_slot (mp, mpp, i)
+		print_map(mpp);
+}
+#endif
+
+static void
+print_mp (struct multipath * mpp)
+{
+	int j, i;
+	struct path * pp = NULL;
+	struct pathgroup * pgp = NULL;
+
+	if (mpp->alias)
+		printf("%s", mpp->alias);
+
+	if (conf->verbosity == 1) {
+		printf("\n");
+		return;
+	}
+	if (strncmp(mpp->alias, mpp->wwid, WWID_SIZE))
+		printf(" (%s)", mpp->wwid);
+
+	printf("\n");
+
+	if (mpp->size)
+		printf("[size=%lu MB]", mpp->size / 2 / 1024);
+
+	if (mpp->features)
+		printf("[features=\"%s\"]", mpp->features);
+
+	if (mpp->hwhandler)
+		printf("[hwhandler=\"%s\"]", mpp->hwhandler);
+
+	fprintf(stdout, "\n");
+
+	if (!mpp->pg)
+		return;
+
+	vector_foreach_slot (mpp->pg, pgp, j) {
+		printf("\\_ ");
+
+		if (mpp->selector)
+			printf("%s ", mpp->selector);
+
+		if (mpp->nextpg && mpp->nextpg == j + 1)
+			printf("[first]");
+
+		switch (pgp->status) {
+		case PGSTATE_ENABLED:
+			printf("[enabled]\n");
+			break;
+		case PGSTATE_DISABLED:
+			printf("[disabled]\n");
+			break;
+		case PGSTATE_ACTIVE:
+			printf("[active]\n");
+			break;
+		default:
+			printf("\n");
+			break;
+		}
+		vector_foreach_slot (pgp->paths, pp, i)
+			print_path(pp, PRINT_PATH_SHORT);
+	}
+	printf("\n");
+}
+
 static void
 print_all_mp (vector mp)
 {
-	int k, i;
+	int k;
 	struct multipath * mpp;
-	struct path * pp = NULL;
 
-	fprintf(stdout, "#\n# all multipaths :\n#\n");
-
-	vector_foreach_slot (mp, mpp, k) {
-		if (mpp->alias)
-			printf("%s (%s)", mpp->alias, mpp->wwid);
-		else
-			printf("%s", mpp->wwid);
-
-		if (!mpp->paths) {
-			printf("\n");
-			continue;
-		}
-		pp = VECTOR_SLOT(mpp->paths, 0);
-
-		if (pp->product_id)
-			printf(" [%.16s]\n", pp->product_id);
-
-		vector_foreach_slot (mpp->paths, pp, i)
-			print_path(pp, SHORT);
-	}
+	vector_foreach_slot (mp, mpp, k) 
+		print_mp(mpp);
 }
 
 static void
@@ -693,7 +351,6 @@ coalesce_paths (vector mp, vector pathvec)
 	int k, i, already_done;
 	char empty_buff[WWID_SIZE];
 	struct multipath * mpp;
-	struct mpentry * mpe;
 	struct path * pp1;
 	struct path * pp2;
 
@@ -724,9 +381,8 @@ coalesce_paths (vector mp, vector pathvec)
 		mpp = zalloc(sizeof(struct multipath));
 		strcpy (mpp->wwid, pp1->wwid);
 
-		mpe = find_mp(mpp->wwid);
-		if (mpe)
-			mpp->alias = mpe->alias;
+		mpp->mpe = find_mp(pp1->wwid);
+		mpp->hwe = find_hw(pp1->vendor_id, pp1->product_id);
 
 		mpp->paths = vector_alloc();
 		vector_alloc_slot (mpp->paths);
@@ -751,42 +407,74 @@ coalesce_paths (vector mp, vector pathvec)
 	}
 }
 
+static int
+dm_switchgroup(char * mapname, int index)
+{
+	int r = 0;
+	struct dm_task *dmt;
+	char str[24];
+
+	if (!(dmt = dm_task_create(DM_DEVICE_TARGET_MSG)))
+		return 0;
+
+	if (!dm_task_set_name(dmt, mapname))
+		goto out;
+
+	if (!dm_task_set_sector(dmt, 0))
+		goto out;
+
+	snprintf(str, 24, "switch_group %i\n", index);
+	dbg("message %s 0 %s", mapname, str);
+
+	if (!dm_task_set_message(dmt, str))
+		goto out;
+
+	if (!dm_task_run(dmt))
+		goto out;
+
+	r = 1;
+
+	out:
+	dm_task_destroy(dmt);
+
+	return r;
+}
 static int
 dm_reinstate(char * mapname, char * path)
 {
-        int r = 0;
+	int r = 0;
 	int sz;
-        struct dm_task *dmt;
-        char *str;
+	struct dm_task *dmt;
+	char *str;
 
-        if (!(dmt = dm_task_create(DM_DEVICE_TARGET_MSG)))
-                return 0;
+	if (!(dmt = dm_task_create(DM_DEVICE_TARGET_MSG)))
+		return 0;
 
-        if (!dm_task_set_name(dmt, mapname))
-                goto out;
+	if (!dm_task_set_name(dmt, mapname))
+		goto out;
 
-        if (!dm_task_set_sector(dmt, 0))
-                goto out;
+	if (!dm_task_set_sector(dmt, 0))
+		goto out;
 
-	sz = strlen(path) + 11;
-        str = zalloc(sz);
+	sz = strlen(path) + 16;
+	str = zalloc(sz);
 
-	snprintf(str, sz, "reinstate %s\n", path);
+	snprintf(str, sz, "reinstate_path %s\n", path);
 
-        if (!dm_task_set_message(dmt, str))
-                goto out;
+	if (!dm_task_set_message(dmt, str))
+		goto out;
 
-        free(str);
+	free(str);
 
-        if (!dm_task_run(dmt))
-                goto out;
+	if (!dm_task_run(dmt))
+		goto out;
 
-        r = 1;
+	r = 1;
 
-      out:
-        dm_task_destroy(dmt);
+	out:
+	dm_task_destroy(dmt);
 
-        return r;
+	return r;
 }
 
 /*
@@ -798,15 +486,15 @@ assemble_map (struct multipath * mp)
 	int i, j;
 	int shift, freechar;
 	char * p;
-	vector pgpaths;
+	struct pathgroup * pgp;
 	struct path * pp;
 
 	p = mp->params;
 	freechar = sizeof(mp->params);
 	
-	shift = snprintf(p, freechar, "%s %s %i 1",
+	shift = snprintf(p, freechar, "%s %s %i %i",
 			 mp->features, mp->hwhandler,
-			 VECTOR_SIZE(mp->pg));
+			 VECTOR_SIZE(mp->pg), mp->nextpg);
 
 	if (shift >= freechar) {
 		fprintf(stderr, "mp->params too small\n");
@@ -815,10 +503,10 @@ assemble_map (struct multipath * mp)
 	p += shift;
 	freechar -= shift;
 	
-	for (i = VECTOR_SIZE(mp->pg) - 1; i >= 0; i--) {
-		pgpaths = VECTOR_SLOT(mp->pg, i);
-		shift = snprintf(p, freechar, " %s %i %i", mp->selector,
-				 VECTOR_SIZE(pgpaths), mp->selector_args);
+	vector_foreach_slot (mp->pg, pgp, i) {
+		pgp = VECTOR_SLOT(mp->pg, i);
+		shift = snprintf(p, freechar, " %s %i 1", mp->selector,
+				 VECTOR_SIZE(pgp->paths));
 		if (shift >= freechar) {
 			fprintf(stderr, "mp->params too small\n");
 			exit(1);
@@ -826,8 +514,8 @@ assemble_map (struct multipath * mp)
 		p += shift;
 		freechar -= shift;
 
-		vector_foreach_slot (pgpaths, pp, j) {
-			shift = snprintf(p, freechar, " %s", pp->dev_t);
+		vector_foreach_slot (pgp->paths, pp, j) {
+			shift = snprintf(p, freechar, " %s 1000" ,pp->dev_t);
 			if (shift >= freechar) {
 				fprintf(stderr, "mp->params too small\n");
 				exit(1);
@@ -844,14 +532,12 @@ assemble_map (struct multipath * mp)
 }
 
 static int
-setup_map (vector pathvec, struct multipath * mpp)
+setup_map (struct multipath * mpp)
 {
 	struct path * pp;
-	int i;
-	char * mapname = NULL;
-	char * curparams = NULL;
-	int op;
-	int r = 0;
+	struct pathgroup * pgp;
+	int i, j;
+	int highest = 0;
 
 	/*
 	 * don't bother if devmap size is unknown
@@ -867,8 +553,6 @@ setup_map (vector pathvec, struct multipath * mpp)
 		if (pp->claimed)
 			return 1;
 
-	pp = VECTOR_SLOT(mpp->paths, 0);
-
 	/*
 	 * properties selectors
 	 */
@@ -876,23 +560,10 @@ setup_map (vector pathvec, struct multipath * mpp)
 	select_selector(mpp);
 	select_features(mpp);
 	select_hwhandler(mpp);
+	select_alias(mpp);
 
 	/*
-	 * layered map computation :
-	 *  1) separate failed paths in a tersary PG
-	 *  2) separate shaky paths in a secondary PG
-	 *  3) apply selected grouping policy to valid paths
-	 *  4) reorder path groups by summed priority
-	 */
-
-	/*
-	 * 1) & 2)
-	 */
-	//group_by_status(mpp, PATH_DOWN);
-	//group_by_status(mpp, PATH_SHAKY);
-
-	/*
-	 * 3) apply selected grouping policy
+	 * apply selected grouping policy to valid paths
 	 */
 	if (mpp->pgpolicy == MULTIBUS)
 		one_group (mpp);
@@ -912,32 +583,156 @@ setup_map (vector pathvec, struct multipath * mpp)
 	}
 
 	/*
-	 * 4)
+	 * ponders each path group and determine highest prio pg
 	 */
-	sort_pg_by_summed_prio(mpp);
+	mpp->nextpg = 1;
+	vector_foreach_slot (mpp->pg, pgp, i) {
+		vector_foreach_slot (pgp->paths, pp, j) {
+			pgp->id ^= (int)pp;
+			if (pp->state != PATH_DOWN)
+				pgp->priority += pp->priority;
+		}
+		if (pgp->priority > highest) {
+			highest = pgp->priority;
+			mpp->nextpg = i + 1;
+		}
+	}
 
 	/*
 	 * transform the mp->pg vector of vectors of paths
 	 * into a mp->params strings to feed the device-mapper
 	 */
 	assemble_map(mpp);
+	return 0;
+}
+
+static int
+pgcmp (struct multipath * mpp, struct multipath * cmpp)
+{
+	int i, j;
+	struct pathgroup * pgp;
+	struct pathgroup * cpgp;
+	int r = 0;
+
+	vector_foreach_slot (mpp->pg, pgp, i) {
+		vector_foreach_slot (cmpp->pg, cpgp, j) {
+			if (pgp->id == cpgp->id) {
+				r = 0;
+				break;
+			}
+			r++;
+		}
+		if (r)
+			return r;
+	}
+	return r;
+}
+
+static int
+select_action (struct multipath * mpp, vector curmp)
+{
+	int i;
+	struct multipath * cmpp;
 
 	/*
-	 * select between RELOAD and CREATE
+	 * FIXME : ACT_RENAME, move to action |= ACT_*
 	 */
-	mapname = mpp->alias ? mpp->alias : mpp->wwid;
-	op = dm_map_present(mapname) ? DM_DEVICE_RELOAD : DM_DEVICE_CREATE;
-	
+
+	vector_foreach_slot (curmp, cmpp, i) {
+		if (strncmp(cmpp->alias, mpp->alias, strlen(mpp->alias)))
+			continue;
+
+		if (cmpp->size != mpp->size) {
+			dbg("size different than current");
+			return ACT_RELOAD;
+		}
+		if (strncmp(cmpp->features, mpp->features,
+			    strlen(mpp->features))) {
+			dbg("features different than current");
+			return ACT_RELOAD;
+		}
+		if (strncmp(cmpp->hwhandler, mpp->hwhandler,
+			    strlen(mpp->hwhandler))) {
+			dbg("hwhandler different than current");
+			return ACT_RELOAD;
+		}
+		if (strncmp(cmpp->selector, mpp->selector,
+			    strlen(mpp->selector))) {
+			dbg("selector different than current");
+			return ACT_RELOAD;
+		}
+		if (VECTOR_SIZE(cmpp->pg) != VECTOR_SIZE(mpp->pg)) {
+			dbg("different path group topology");
+			return ACT_RELOAD;
+		}
+		if (pgcmp(mpp, cmpp)) {
+			dbg("different path group topology");
+			return ACT_RELOAD;
+		}
+		if (cmpp->nextpg != mpp->nextpg) {
+			dbg("nextpg different than current");
+			return ACT_SWITCHPG;
+		}
+		return ACT_NOTHING;
+	}
+	return ACT_CREATE;
+}
+
+static int
+reinstate_paths (struct multipath * mpp)
+{
+	int i, j;
+	struct pathgroup * pgp;
+	struct path * pp;
+
+	vector_foreach_slot (mpp->pg, pgp, i) {
+		if (pgp->status == PGSTATE_DISABLED ||
+		    pgp->status == PGSTATE_ACTIVE)
+			continue;
+
+		vector_foreach_slot (pgp->paths, pp, j) {
+			if (pp->dmstate == PSTATE_FAILED) {
+				dm_reinstate(mpp->alias, pp->dev_t);
+			}
+		}
+	}
+	return 0;
+}
+
+static int
+domap (struct multipath * mpp, int action)
+{
+	int op;
+	int r = 0;
+	char * act;
+
+	switch (action) {
+	case ACT_RELOAD:
+		op = DM_DEVICE_RELOAD;
+		act = ACT_RELOAD_STR;
+		break;
+
+	case ACT_CREATE:
+		op = DM_DEVICE_CREATE;
+		act = ACT_CREATE_STR;
+		break;
+
+	case ACT_SWITCHPG:
+		act = ACT_SWITCHPG_STR;
+		break;
+
+	case ACT_NOTHING:
+		act = ACT_NOTHING_STR;
+		break;
+
+	default:
+		break;
+	}
+
 	if (conf->verbosity > 1)
-		printf("%s:", (op == DM_DEVICE_RELOAD) ? "reload" : "create");
-	if (conf->verbosity > 0)
-		printf("%s", mapname);
-	if (conf->verbosity > 1)
-		printf(":0 %lu %s %s",
-			mpp->size, DM_TARGET,
-			mpp->params);
-	if (conf->verbosity > 0)
-		printf("\n");
+		printf("%s: ", act);
+
+	print_mp(mpp);
 
 	/*
 	 * last chance to quit before touching the devmaps
@@ -945,6 +740,14 @@ setup_map (vector pathvec, struct multipath * mpp)
 	if (conf->dry_run)
 		return 0;
 
+	if (action == ACT_NOTHING)
+		return 0;
+
+	if (action == ACT_SWITCHPG) {
+		dm_switchgroup(mpp->alias, mpp->nextpg);
+		return 0;
+	}
+		
 	/*
 	 * device mapper creation or updating
 	 * here we know we'll have garbage on stderr from
@@ -953,40 +756,21 @@ setup_map (vector pathvec, struct multipath * mpp)
 	dm_log_init_verbose(0);
 
 	if (op == DM_DEVICE_RELOAD) {
-		if (conf->dev && filepresent(conf->dev) &&
-		    dm_get_map(mapname, NULL, &curparams) &&
-		    0 == strncmp(mpp->params, curparams, strlen(mpp->params))) {
-	                pp = zalloc(sizeof(struct path));
-	                basename(conf->dev, pp->dev);
-
-	                if (devinfo(pp)) {
-				r = 1;
-	                        goto out;
-			}
-			dm_reinstate(mapname, pp->dev_t);
-	                free(pp);
-
-			if (conf->verbosity > 1)
-				printf("[reinstate %s in %s]\n",
-					conf->dev, mapname);
-			goto out;
-		}
-		if (!dm_simplecmd(DM_DEVICE_SUSPEND, mapname))
+		if (!dm_simplecmd(DM_DEVICE_SUSPEND, mpp->alias))
 			goto out;
 	}
 
-	if (!dm_addmap(op, mapname, DM_TARGET, mpp->params, mpp->size)) {
-		dm_simplecmd(DM_DEVICE_REMOVE, mapname);
+	if (!dm_addmap(op, mpp->alias, DM_TARGET, mpp->params, mpp->size)) {
+		dm_simplecmd(DM_DEVICE_REMOVE, mpp->alias);
 		goto out;
 	}
 
 	if (op == DM_DEVICE_RELOAD)
-		if (!dm_simplecmd(DM_DEVICE_RESUME, mapname))
+		if (!dm_simplecmd(DM_DEVICE_RESUME, mpp->alias))
 			goto out;
 
 	out:
 	dm_log_init_verbose(1);
-	free(curparams);
 	return r;
 }
 
@@ -1000,6 +784,7 @@ dm_get_maps (vector mp, char * type)
 	unsigned next = 0;
 	unsigned long length;
 	char *params;
+	char *status;
 
 	if (!(dmt = dm_task_create (DM_DEVICE_LIST)))
 		return 0;
@@ -1016,6 +801,7 @@ dm_get_maps (vector mp, char * type)
 	do {
 		if (dm_type(names->name, DM_TARGET)) {
 			dm_get_map(names->name, &length, &params);
+			dm_get_status(names->name, &status);
 			mpp = zalloc(sizeof(struct multipath));
 
 			if (!mpp) {
@@ -1023,8 +809,10 @@ dm_get_maps (vector mp, char * type)
 				goto out;
 			}
 			mpp->size = length;
-			strncat(mpp->wwid, names->name, WWID_SIZE);
+			mpp->alias = zalloc(strlen(names->name) + 1);
+			strncat(mpp->alias, names->name, strlen(names->name));
 			strncat(mpp->params, params, PARAMS_SIZE);
+			strncat(mpp->status, status, PARAMS_SIZE);
 
 			vector_alloc_slot(mp);
 			vector_set_slot(mp, mpp);
@@ -1307,7 +1095,7 @@ main (int argc, char *argv[])
 	}
 
 	/*
-	 * allocate the two core vectors to store paths and multipaths
+	 * allocate core vectors to store paths and multipaths
 	 */
 	mp = vector_alloc();
 	curmp = vector_alloc();
@@ -1318,12 +1106,6 @@ main (int argc, char *argv[])
 		exit(1);
 	}
 
-	if (conf->list) {
-		dm_get_maps(curmp, DM_TARGET);
-		print_all_mp(curmp);
-		goto out;
-	}
-
 	/*
 	 * read the config file
 	 */
@@ -1365,11 +1147,25 @@ main (int argc, char *argv[])
 	if (get_pathvec_sysfs(pathvec))
 		exit(1);
 
+	dm_get_maps(curmp, DM_TARGET);
+
 	if (VECTOR_SIZE(pathvec) == 0 && conf->verbosity > 0) {
 		fprintf(stdout, "no path found\n");
 		exit(0);
 	}
 
+	vector_foreach_slot (curmp, mpp, k) {
+		dbg("params = %s", mpp->params);
+		dbg("status = %s", mpp->status);
+		disassemble_map(pathvec, mpp->params, mpp);
+		disassemble_status(mpp->status, mpp);
+		reinstate_paths(mpp);
+	}
+	if (conf->list) {
+		print_all_mp(curmp);
+		goto out;
+	}
+
 	coalesce_paths(mp, pathvec);
 
 	/*
@@ -1381,22 +1177,23 @@ main (int argc, char *argv[])
 		    (mpp->alias &&
 		    0 == strncmp(mpp->alias, conf->dev, FILE_NAME_SIZE)) ||
 		    0 == strncmp(mpp->wwid, conf->dev, FILE_NAME_SIZE))) {
-			setup_map(pathvec, mpp);
+			setup_map(mpp);
+			domap(mpp, select_action(mpp, curmp));
 			goto out;
 		}
 	}
 
-	if (conf->verbosity > 1) {
-		print_all_path(pathvec);
-		print_all_mp(mp);
+	vector_foreach_slot (mp, mpp, k) {
+		setup_map(mpp);
+		domap(mpp, select_action(mpp, curmp));
 	}
 
-	if (conf->verbosity > 1)
-		fprintf(stdout, "#\n# device maps :\n#\n");
-
-	vector_foreach_slot (mp, mpp, k)
-		setup_map(pathvec, mpp);
-
+#if DEBUG
+	fprintf(stdout, "#\n# all paths :\n#\n");
+	print_all_paths(pathvec);
+	fprintf(stdout, "#\n# device maps :\n#\n");
+	print_all_maps(mp);
+#endif
 out:
 	/*
 	 * signal multipathd that new devmaps may have come up
