@@ -20,21 +20,11 @@ int get_pgpolicy_id(char *);
 void get_pgpolicy_name (char *, int);
 
 /*
- * intermediate policies
- */
-void group_by_status(struct multipath *, int);
-
-/*
- * leaf policies
+ * policies
  */
 void one_path_per_group(struct multipath *);
 void one_group(struct multipath *);
 void group_by_serial(struct multipath *);
 void group_by_prio(struct multipath *);
 
-/*
- * path group reordering policies
- */
-void sort_pg_by_summed_prio(struct multipath *);
-
 #endif
